program Oper

  use sep
  use efwi_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, gpu_id
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: h_nx, h_ox
  integer                             :: bc1, bc2, bc3, bc4
  real                                :: dt, d1, o1, d2, o2

  logical                             :: adj, linear

  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:,:), allocatable :: vel0, deltaV, data

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)
  call from_history("n3", h_nx, 1)
  call from_history("o3", h_ox, 0)

  ! Read wavelet information
  call from_aux("wavelet", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)

  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_id", gpu_id)

  call from_param("adj", adj)
  if(.not. adj) call from_param("linear", linear)

  ! Allocation
  allocate( wavelet(nt))
  allocate( vel0(n1, n2, h_nx))
  allocate( data(nt, rec_nx, sou_nx))
  allocate( deltaV(n1, n2, h_nx))

  wavelet = 0.
  vel0 = 0.
  data = 0.
  deltaV = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")

  ! Call L_initiate
  call L_init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call L_init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call L_init3(h_nx, h_ox, nt, dt, wfld_j, gpu_id)
  call L_init4(wavelet)
  call L_model_update(vel0)

  if(adj) then
    call to_history("n1", n1)
    call to_history("o1", o1)
    call to_history("d1", d1)
    call to_history("n2", n2)
    call to_history("o2", o2)
    call to_history("d2", d2)
    call to_history("n3", h_nx)
    call to_history("o3", h_ox)
    call to_history("d3", 1.)

    call sep_read(data, "data")
    call L_oper2(adj, .false., deltaV, data)
    call sep_write(deltaV)
  else
    call to_history("n1", nt)
    call to_history("o1", 0.)
    call to_history("d1", dt)
    call to_history("n2", rec_nx)
    call to_history("o2", o2+d2*(rec_ox-1))
    call to_history("d2", d2*rec_dx)
    call to_history("n3", sou_nx)
    call to_history("o3", o2+d2*(sou_ox-1))
    call to_history("d3", d2*sou_dx)

    if(linear) then
      call sep_read(deltaV, "deltaV")
      call L_oper2(adj, .false., deltaV, data)
    else
      call nlL_oper2(.false., data)
    endif
    call sep_write(data)
  endif

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
