program Oper

  use sep
  use efwi_dso_mod
  use tomo_dso_mod
  use mute_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, gpu_id
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: h_nx, h_ox, ih, ihx
  integer                             :: bc1, bc2, bc3, bc4
  real                                :: dt, d1, o1, d2, o2

  logical                             :: dso, mute
  real                                :: eps
  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:,:,:), allocatable :: data_obs, data, vel_init, vel0, vel1, gradient, hwt

  integer                             :: stat, iter, niter, itry, ntry, counter

  double precision                    :: rd0, rd1, rd2, rd3
  double precision                    :: rm0, rm1, rm2, rm3
  double precision                    :: r0, r1, r2, r3
  double precision                    :: step_size

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)
  call from_history("n3", h_nx)
  call from_history("o3", h_ox)

  ! Read wavelet information
  call from_aux("wavelet", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)

  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_id", gpu_id)

  call from_param('niter', niter, 10)
  call from_param('ntry', ntry, 2)
  call from_param('dso', dso)
  call from_param('mute', mute)

  if(dso) then
    call from_param('eps', eps)
    if(eps<0.) call erexit("eps value is not correct")
  endif

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", h_nx)
  call to_history("o3", h_ox)
  call to_history("d3", 1.)

  call to_history("n1", n1, "mmov")
  call to_history("o1", o1, "mmov")
  call to_history("d1", d1, "mmov")
  call to_history("n2", n2, "mmov")
  call to_history("o2", o2, "mmov")
  call to_history("d2", d2, "mmov")
  call to_history("n3", h_nx, "mmov")
  call to_history("o3", h_ox, "mmov")
  call to_history("d3", 1., "mmov")

  call to_history("n1", 3, "res")
  call to_history("o1", 0., "res")
  call to_history("d1", 1., "res")

  ! Allocation
  allocate( wavelet(nt))
  allocate( res(3))
  allocate( vel_init(n1, n2, h_nx))
  allocate( vel0(n1, n2, h_nx))
  allocate( vel1(n1, n2, h_nx))
  allocate( data_obs(nt, rec_nx, sou_nx))
  allocate( data(nt, rec_nx, sou_nx))
  allocate( gradient(n1, n2, h_nx))
  if(dso) allocate( hwt(n1, n2, h_nx))

  wavelet = 0.
  vel_init = 0.
  vel0 = 0.
  data_obs = 0.
  res = 0.
  rm0 = 0.
  rm1 = 0.
  rm2 = 0.
  rm3 = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")
  call sep_read(data_obs, "data")

  vel_init = vel0

  ! Call L_initiate
  call L_init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call L_init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call L_init3(h_nx, h_ox, nt, dt, wfld_j, gpu_id)
  call L_init4(wavelet)
  call L_model_update(vel0)

  if(dso) then
    do ih=1,h_nx
      hwt(:,:,ih) = ih-1+h_ox
    enddo
    hwt = abs(hwt)
    if(maxval(hwt) > 0.) hwt = eps*hwt/maxval(hwt)
  endif

  if(mute) then
    call mute_init(nt, sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx)
    if(mute) call mute_offset(data_obs)
  endif

  counter = 0

  do iter=1,niter
    call L_model_update(vel0)
    call nlL_oper2(.false., data)
    if(mute) call mute_offset(data)
    data = data - data_obs
    rd0 = sum(dprod(data,data))
    if(dso) rm0 = sum(dprod(hwt*vel0,hwt*vel0))
    r0 = rd0 + rm0

    call L_oper2(.true., .false., gradient, data)

    if(iter .eq. 1) then
      step_size = minval(vel0(:,:,1-h_ox)) / maxval(abs(gradient)) / 100.
      write(0,*) "initial step size is =", step_size
    endif

    do itry=1,2*ntry

      if(itry .eq. ntry) then
        step_size = minval(vel0(:,:,1-h_ox)) / maxval(abs(gradient)) / 100.
        write(0,*) "Step size is resetted to ", step_size
      endif

      vel1 = vel0 - gradient * step_size * 0.5
      call L_model_update(vel1)
      call nlL_oper2(.false., data)
      if(mute) call mute_offset(data)
      data = data - data_obs
      rd1 = sum(dprod(data,data))
      if(dso) rm1 = sum(dprod(hwt*vel1,hwt*vel1))
      r1 = rd1 + rm1

      vel1 = vel0 - gradient * step_size * 1.0
      call L_model_update(vel1)
      call nlL_oper2(.false., data)
      if(mute) call mute_offset(data)
      data = data - data_obs
      rd2 = sum(dprod(data,data))
      if(dso) rm2 = sum(dprod(hwt*vel1,hwt*vel1))
      r2 = rd2 + rm2

      vel1 = vel0 - gradient * step_size * 1.5
      call L_model_update(vel1)
      call nlL_oper2(.false., data)
      if(mute) call mute_offset(data)
      data = data - data_obs
      rd3 = sum(dprod(data,data))
      if(dso) rm3 = sum(dprod(hwt*vel1,hwt*vel1))
      r3 = rd3 + rm3

      write(0,*) "data  residuals =", real(rd0), real(rd1), real(rd2), real(rd3)
      write(0,*) "model residuals =", real(rm0), real(rm1), real(rm2), real(rm3)
      write(0,*) "total residuals =", real(r0), real(r1), real(r2), real(r3)

      call stepper(rd0, rd1, rd2, rd3, step_size, stat)

      if(stat .eq. 0) exit

    end do ! end of trying loop

    if(stat .ne. 0) then
      write(0,*) "trying iterations reached limit without finding a proper step size, will exit the program"
      exit
    endif

    write(0,*) " "

    vel0 = vel0 - step_size * gradient
    if(dso) vel0 = vel0 - hwt * hwt * vel0
    res(1) = rd0
    res(2) = rm0
    res(3) = r0

    ! Output model
    call sep_write(vel0-vel_init, "mmov")
    call sep_write(res, "res")

    counter = counter + 1
    call to_history("n4", counter, "mmov")
    call to_history("n2", counter, "res")

  end do ! end of iteration loop

  call sep_write(vel0)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
