program Oper

  use sep
  use efwi_pe_mod
  use tomo_pe_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, gpu_id
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: h_nx, h_ox
  integer                             :: bc1, bc2, bc3, bc4
  real                                :: dt, d1, o1, d2, o2

  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:,:), allocatable   :: wavelet_rands, data, data_obs_rands
  real, dimension(:,:,:), allocatable :: data_obs, vel0, vel1, gradient

  integer                             :: stat, iter, niter, itry, ntry, counter

  double precision                    :: r0, r1, r2, r3, step_size

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)
  call from_history("n3", h_nx)
  call from_history("o3", h_ox)

  ! Read wavelet information
  call from_aux("wavelet", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)

  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_id", gpu_id)

  call from_param('niter', niter, 10)
  call from_param('ntry', ntry, 2)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", h_nx)
  call to_history("o3", h_ox)
  call to_history("d3", 1.)

  call to_history("n1", n1, "mmov")
  call to_history("o1", o1, "mmov")
  call to_history("d1", d1, "mmov")
  call to_history("n2", n2, "mmov")
  call to_history("o2", o2, "mmov")
  call to_history("d2", d2, "mmov")
  call to_history("n3", h_nx, "mmov")
  call to_history("o3", h_ox, "mmov")
  call to_history("d3", 1., "mmov")

  call to_history("n1", 3, "res")
  call to_history("o1", 0., "res")
  call to_history("d1", 1., "res")

  ! Allocation
  allocate( wavelet(nt))
  allocate( res(3))
  allocate( wavelet_rands(nt, sou_nx))
  allocate( vel0(n1, n2, h_nx))
  allocate( vel1(n1, n2, h_nx))
  allocate( data(nt, rec_nx))
  allocate( data_obs(nt, rec_nx, sou_nx))
  allocate( data_obs_rands(nt, rec_nx))
  allocate( gradient(n1, n2, h_nx))

  wavelet = 0.
  vel0 = 0.
  data_obs = 0.
  res = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")
  call sep_read(data_obs, "data")

  ! Call L_initiate
  call L_init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call L_init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call L_init3(h_nx, h_ox, nt, dt, wfld_j, gpu_id)
  call L_init4()
  call L_model_update(vel0)

  counter = 0

  do iter=1,niter*sou_nx
    call L_encode(wavelet, wavelet_rands, data_obs, data_obs_rands)
    call L_model_update(vel0)
    call nlL_oper2(.false., data)
    data = data_obs_rands - data
    r0 = sum(dprod(data,data))
    if(r0 .eq. 0.0) then
      write(0,*) "r0 is zero, will exit the program"
      exit
    end if
    call L_oper2(.true., .false., gradient, data)
    !write(0,*) "gradient stats", minval(gradient), maxval(gradient), sum(gradient)
    if(iter .eq. 1) then
      step_size = minval(vel0(:,:,1-h_ox)) / maxval(abs(gradient)) / 100.
      write(0,*) "initial step size is =", step_size
    endif
    do itry=1,2*ntry

      if(itry .eq. ntry) then
        step_size = minval(vel0(:,:,1-h_ox)) / maxval(abs(gradient)) / 100.
        write(0,*) "Step size is resetted to ", step_size
      endif

      vel1 = vel0 + gradient * step_size * 0.5
      call L_model_update(vel1)
      call nlL_oper2(.false., data)
      data = data_obs_rands - data
      r1 = sum(dprod(data,data))

      vel1 = vel0 + gradient * step_size * 1.0
      call L_model_update(vel1)
      call nlL_oper2(.false., data)
      data = data_obs_rands - data
      r2 = sum(dprod(data,data))

      vel1 = vel0 + gradient * step_size * 1.5
      call L_model_update(vel1)
      call nlL_oper2(.false., data)
      data = data_obs_rands - data
      r3 = sum(dprod(data,data))

      write(0,*) "residuals =", real(r0), real(r1), real(r2), real(r3)
      call stepper(r0, r1, r2, r3, step_size, stat)

      if(stat .eq. 0) exit

    end do ! end of trying loop

    if(stat .ne. 0) then
      write(0,*) "trying iterations reached limit without finding a proper step size, will exit the program"
      exit
    endif

    write(0,*) " "
    gradient = step_size * gradient
    vel0 = vel0 + gradient
    ! Output model
    if(mod(iter, sou_nx) .eq. 0) then
      res(1) = r0
      call sep_write(vel0, "mmov")
      call sep_write(res, "res")
      counter = counter + 1
      call to_history("n4", counter, "mmov")
      call to_history("n2", counter, "res")
    endif
  end do ! end of iteration loop

  call sep_write(vel0)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
