program Stepper

  use sep

  implicit none

  integer                             :: n1m, n2m, n3m
  integer                             :: n1d, n2d, n3d
  real                                :: o1d, d1d, o2d, d2d, o3d, d3d
  real                                :: o1m, d1m, o2m, d2m, o3m, d3m

  real, dimension(:,:,:), allocatable :: model, grad
  real, dimension(:,:,:), allocatable :: res, grad_data

  double precision                    :: alfa

  call sep_init()

  call from_history("n1m", n1m)
  call from_history("o1m", o1m)
  call from_history("d1m", d1m)
  call from_history("n2m", n2m)
  call from_history("o2m", o2m)
  call from_history("d2m", d2m)
  call from_history("n3m", n3m)
  call from_history("o3m", o3m)
  call from_history("d3m", d3m)

  call from_aux("res", "n1", n1d)
  call from_aux("res", "o1", o1d)
  call from_aux("res", "d1", d1d)
  call from_aux("res", "n2", n2d)
  call from_aux("res", "o2", o2d)
  call from_aux("res", "d2", d2d)
  call from_aux("res", "n3", n3d)
  call from_aux("res", "o3", o3d)
  call from_aux("res", "d3", d3d)

  call to_history("n1", n1m)
  call to_history("o1", o1m)
  call to_history("d1", d1m)
  call to_history("n2", n2m)
  call to_history("o2", o2m)
  call to_history("d2", d2m)
  call to_history("n3", n3m)
  call to_history("o3", o3m)
  call to_history("d3", d3m)

  call to_history("n1", n1d, "updated_res")
  call to_history("o1", o1d, "updated_res")
  call to_history("d1", d1d, "updated_res")
  call to_history("n2", n2d, "updated_res")
  call to_history("o2", o2d, "updated_res")
  call to_history("d2", d2d, "updated_res")
  call to_history("n3", n3d, "updated_res")
  call to_history("o3", o3d, "updated_res")
  call to_history("d3", d3d, "updated_res")

  allocate(model(n1m,n2m,n3m))
  allocate(grad(n1m,n2m,n3m))
  allocate(res(n1d,n2d,n3d))
  allocate(grad_data(n1d,n2d,n3d))

  call sep_read(model)
  call sep_read(res,"res")
  call sep_read(grad,"grad")
  call sep_read(grad_data,"grad_data")

  alfa = - sum(dprod(grad_data,res)) / sum(dprod(grad_data,grad_data))
  if(isnan(alfa)) call erexit("alfa is NaN")

  model = model + alfa * grad
  res = res + alfa * grad_data

  call sep_write(model)
  call sep_write(res,"updated_res")

  call sep_close()

end program
