module efwi_dso_mod

  use tomo_dso_mod

  implicit none
  integer, private                             :: nt, n1, n2, n3, wfld_j, gpu_id
  integer, private                             :: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x
  integer, private                             :: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x
  integer, private                             :: h_nx, h_ox, ih, h_x, ix
  integer, private                             :: n1e, n2e, bc1, bc2, bc3, bc4, fat, iz
  real, private                                :: dt, d1, o1, d2, o2, dtw

  real, dimension(:), allocatable, private     :: wavelet
  real, dimension(:,:), allocatable, private   :: vwt, wt
  real, dimension(:,:,:), allocatable, private :: vel0e, temp_deltaV, temp_data
  contains

  subroutine L_init1(n1_in, o1_in, d1_in, n2_in, o2_in, d2_in, bc1_in, bc2_in, bc3_in, bc4_in)
    integer                             :: n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                                :: d1_in, o1_in, d2_in, o2_in
    n1 = n1_in
    n2 = n2_in
    d1 = d1_in
    o1 = o1_in
    d2 = d2_in
    o2 = o2_in
    bc1 = bc1_in
    bc2 = bc2_in
    bc3 = bc3_in
    bc4 = bc4_in
  end subroutine

  subroutine L_init2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in)
    integer                             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
    integer                             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in
    sou_nx = sou_nx_in
    sou_ox = sou_ox_in
    sou_dx = sou_dx_in
    sou_z = sou_z_in
    rec_nx = rec_nx_in
    rec_ox = rec_ox_in
    rec_dx = rec_dx_in
    rec_z = rec_z_in
  end subroutine

  subroutine L_init3(h_nx_in, h_ox_in, nt_in, dt_in, wfld_j_in, gpu_id_in)
    integer                             :: h_nx_in, h_ox_in, nt_in, wfld_j_in, gpu_id_in
    real                                :: dt_in
    h_nx = h_nx_in
    h_ox = h_ox_in
    nt = nt_in
    wfld_j = wfld_j_in
    gpu_id = gpu_id_in
    dt = dt_in
  end subroutine

  subroutine L_init4(wavelet_in)
    real, dimension(nt)                  :: wavelet_in

    fat = 5
    n1e = n1 + bc1 + bc3 + 2*fat
    n2e = n2 + bc2 + bc4 + 2*fat
    n3 = (nt-1)*wfld_j+1
    dtw = dt / wfld_j

    allocate( wavelet(nt))
    allocate( vwt(n1e, n2e), wt(n1e, n2e))
    allocate( vel0e(n1e, n2e, h_nx))
    allocate( temp_data(nt, rec_nx, sou_nx))
    allocate( temp_deltaV(n1, n2, h_nx))

    wavelet = 0.
    vwt = 0.
    wt = 0.

    wavelet = wavelet_in

    call c_setdevice(gpu_id)
    call c_init1(nt, n1, n2, n3, dt, d1, d2)
    call c_init2(h_nx, h_ox, n1e, n2e, bc1, bc2, wfld_j)
    call c_init3(sou_nx, sou_ox-1, sou_dx, sou_z-1, rec_nx, rec_ox-1, rec_dx, rec_z-1)
    call c_alloc(wavelet)

  end subroutine

  subroutine L_model_update(vel0_in)
    real, dimension(n1, n2, h_nx)        :: vel0_in

    vel0e = 0.
    vel0e(1+fat+bc1:n1+fat+bc1,1+fat+bc2:n2+fat+bc2,:) = vel0_in
    do ih=1,h_nx
      call extend(vel0e(:,:,ih), bc1+fat, bc2+fat, bc3+fat, bc4+fat)
    enddo

    if(1-h_ox<1 .or. 1-h_ox>h_nx) call erexit("vel at h=0 doesn't exist")
    vel0e(:,:,1-h_ox) = sqrt(vel0e(:,:,1-h_ox))
    call AbcWt(vel0e(:,:,1-h_ox), vwt, wt, dtw, d1, d2, bc1, bc2, bc3, bc4, fat)
    vel0e(:,:,1-h_ox) = vel0e(:,:,1-h_ox)*vel0e(:,:,1-h_ox)
    call c_model_update(vel0e*dtw*dtw, vwt, wt)

  end subroutine

  function L_oper(adj, add, m, d) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj, add 
    real,dimension(:)  :: m,d
    call L_oper2(adj,add,m,d)
    stat=0
  end function

  function nlL_oper(add, d) result(stat)
    integer            :: stat
    logical,intent(in) :: add 
    real,dimension(:)  :: d
    call nlL_oper2(add,d)
    stat=0
  end function

  subroutine L_oper2(adj, add, deltaV, data)
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2, h_nx)        :: deltaV
    real, dimension(nt, rec_nx, sou_nx)  :: data

    if(adj) then

      if(.not. add) deltaV = 0.
      temp_deltaV = 0.
      call c_migrate(0, 1, temp_deltaV, data)
      deltaV = deltaV + temp_deltaV*wfld_j*wfld_j

    else

      if(.not. add) data = 0.
      temp_data = 0.
      call c_model(0, 1, deltaV, temp_data)
      data = data + temp_data*wfld_j*wfld_j

    endif

  end subroutine

  subroutine nlL_oper2(add, data)
    logical, intent(in)                  :: add
    real, dimension(nt, rec_nx, sou_nx)  :: data

    if(.not. add) data = 0.
    temp_data = 0.
    call c_nlmodel(0, 1, temp_data)
    data = data + temp_data*wfld_j

  end subroutine

end module 
