module mute_mod

  implicit none

  integer, private                                :: nt, offset0, offset1
  integer, private                                :: sou_nx, sou_ox, sou_dx
  integer, private                                :: rec_nx, rec_ox, rec_dx
  integer, private                                :: isou, irec, sou_x, rec_x
  integer, private, dimension(:,:), allocatable   :: offset
  real, private, dimension(:,:), allocatable      :: wt

  contains

  subroutine mute_init(nt_in, sou_nx_in, sou_ox_in, sou_dx_in, rec_nx_in, rec_ox_in, rec_dx_in)

    integer               :: nt_in
    integer               :: sou_nx_in, sou_ox_in, sou_dx_in
    integer               :: rec_nx_in, rec_ox_in, rec_dx_in

    nt = nt_in
    sou_nx = sou_nx_in
    sou_ox = sou_ox_in
    sou_dx = sou_dx_in
    rec_nx = rec_nx_in
    rec_ox = rec_ox_in
    rec_dx = rec_dx_in

    allocate(offset(rec_nx,sou_nx))
    allocate(wt(rec_nx,sou_nx))

    do isou=1,sou_nx
      sou_x = sou_ox + (isou-1)*sou_dx
      do irec=1,rec_nx
        rec_x = rec_ox + (irec-1)*rec_dx
        offset(irec,isou) = rec_x - sou_x
      enddo
    end do

    offset = abs(offset)
    offset1 = maxval(offset(:,(sou_nx+1)/2))
    offset0 = offset1/2

    wt = 1.

    do isou=1,sou_nx
      do irec=1,rec_nx
        if(offset(irec,isou) > offset1) then
          wt(irec,isou) = 0.
        elseif(offset(irec,isou) > offset0 .and. offset(irec,isou) <= offset1 ) then
          wt(irec,isou) = (1.*offset1-offset(irec,isou))/(1.*offset1-offset0)
        endif
      enddo
    end do

  end subroutine

  subroutine mute_offset(data)
    real, dimension(nt, rec_nx, sou_nx)  :: data

    do isou=1,sou_nx
      do irec=1,rec_nx
        data(:,irec,isou) = data(:,irec,isou) * wt(irec,isou)
      enddo
    end do

  end subroutine

end module 
